@@ -3,16 +3,18 @@ package main
 import (
 	"bytes"
 	"crypto/tls"
+	_ "embed"
 	"encoding/json"
 	"errors"
 	"flag"
 	"fmt"
-	"io/ioutil"
+	"html/template"
+	"io"
 	"net"
 	"net/http"
 	"net/url"
 	"os"
-	"runtime"
+	"sort"
 	"strconv"
 	"strings"
 	"sync"
@@ -42,14 +44,17 @@ var (
 	version    = ""
 	buildinfos = ""
 	AppVersion = "RTRmon " + version + " " + buildinfos
+	//go:embed index.html.tmpl
+	IndexTemplate string
 
 	OneOff      = flag.Bool("oneoff", false, "dump as json and exits")
-	Addr        = flag.String("addr", ":8080", "Server address")
+	Addr        = flag.String("addr", ":9866", "Server address")
 	MetricsPath = flag.String("metrics", "/metrics", "Metrics path")
 	OutFile     = flag.String("file", "diff.json", "Diff file (or URL path without /)")
 
 	UserAgent                  = flag.String("useragent", fmt.Sprintf("StayRTR-%v (+https://github.com/bgp/stayrtr)", AppVersion), "User-Agent header")
 	DisableConditionalRequests = flag.Bool("disable.conditional.requests", false, "Disable conditional requests (using If-None-Match/If-Modified-Since headers)")
+	GracePeriod                = flag.Duration("grace.period", 1*time.Hour, "Grace period during which objects removed from a source are not considered for the diff")
 
 	PrimaryHost            = flag.String("primary.host", "tcp://rtr.rpki.cloudflare.com:8282", "primary server")
 	PrimaryValidateCert    = flag.Bool("primary.tls.validate", true, "Validate TLS")
@@ -90,7 +95,7 @@ var (
 	VRPCount = prometheus.NewGaugeVec(
 		prometheus.GaugeOpts{
 			Name: "rpki_vrps",
-			Help: "Total number of VRPS/amount of differents.",
+			Help: "Total number of current VRPS in primary/secondary and current difference between primary and secondary.",
 		},
 		[]string{"server", "url", "type"},
 	)
@@ -101,6 +106,13 @@ var (
 		},
 		[]string{"lhs_url", "rhs_url", "visibility_seconds"},
 	)
+	VRPInGracePeriod = prometheus.NewGaugeVec(
+		prometheus.GaugeOpts{
+			Name: "rpki_grace_period_vrps",
+			Help: "Number of unique VRPS in grace period by url.",
+		},
+		[]string{"url"},
+	)
 	RTRState = prometheus.NewGaugeVec(
 		prometheus.GaugeOpts{
 			Name: "rtr_state",
@@ -142,6 +154,7 @@ var (
 func init() {
 	prometheus.MustRegister(VRPCount)
 	prometheus.MustRegister(VRPDifferenceForDuration)
+	prometheus.MustRegister(VRPInGracePeriod)
 	prometheus.MustRegister(RTRState)
 	prometheus.MustRegister(RTRSerial)
 	prometheus.MustRegister(RTRSession)
@@ -155,6 +168,7 @@ func (t *thresholds) String() string {
 	res := []byte("")
 	for idx, tr := range *t {
 		res = strconv.AppendInt(res, tr, 10)
+
 		if idx < len(*t)-1 {
 			res = append(res, ","...)
 		}
@@ -178,6 +192,9 @@ func (t *thresholds) Set(value string) error {
 		*t = append(*t, threshold)
 	}
 
+	// Sort the breaks in ascending order.
+	sort.Slice(*t, func(i, j int) bool { return (*t)[i] < (*t)[j] })
+
 	return nil
 }
 
@@ -213,9 +230,9 @@ type Client struct {
 	lastUpdate time.Time
 
 	compLock    *sync.RWMutex
-	vrps        map[string]*VRPJsonSimple
+	vrps        VRPMap
 	compRtrLock *sync.RWMutex
-	vrpsRtr     map[string]*VRPJsonSimple
+	vrpsRtr     VRPMap
 
 	unlock chan bool
 	ch     chan int
@@ -229,9 +246,9 @@ type Client struct {
 func NewClient() *Client {
 	return &Client{
 		compLock:    &sync.RWMutex{},
-		vrps:        make(map[string]*VRPJsonSimple),
+		vrps:        make(VRPMap),
 		compRtrLock: &sync.RWMutex{},
-		vrpsRtr:     make(map[string]*VRPJsonSimple),
+		vrpsRtr:     make(VRPMap),
 	}
 }
 
@@ -245,15 +262,13 @@ func (c *Client) Start(id int, ch chan int) {
 	}
 
 	connType := pathUrl.Scheme
-	rtrAddr := fmt.Sprintf("%s", pathUrl.Host)
+	rtrAddr := pathUrl.Host
 
 	bypass := true
 	for {
 
 		if !bypass {
-			select {
-			case <-time.After(c.RefreshInterval):
-			}
+			<-time.After(c.RefreshInterval)
 		}
 		bypass = false
 
@@ -276,7 +291,7 @@ func (c *Client) Start(id int, ch chan int) {
 					serverKeyHash := ssh.FingerprintSHA256(key)
 					if c.ValidateSSH {
 						if serverKeyHash != fmt.Sprintf("SHA256:%v", c.SSHServerKey) {
-							return errors.New(fmt.Sprintf("Server key hash %v is different than expected key hash SHA256:%v", serverKeyHash, c.SSHServerKey))
+							return fmt.Errorf("server key hash %v is different than expected key hash SHA256:%v", serverKeyHash, c.SSHServerKey)
 						}
 					}
 					log.Infof("%d: Connected to server %v via ssh. Fingerprint: %v", id, remote.String(), serverKeyHash)
@@ -307,60 +322,37 @@ func (c *Client) Start(id int, ch chan int) {
 				log.Fatal(err)
 			}
 
-			select {
-			case <-c.qrtr:
-				log.Infof("%d: Quitting RTR session", id)
-			}
+			<-c.qrtr
+			log.Infof("%d: Quitting RTR session", id)
 		} else {
 			log.Infof("%d: Fetching %s", c.id, c.Path)
-			data, _, _, err := c.FetchConfig.FetchFile(c.Path)
-			if err != nil {
-				log.Error(err)
-				continue
-			}
-			log.Debug(data)
-			decoded, err := decodeJSON(data)
-			if err != nil {
+			data, statusCode, _, err := c.FetchConfig.FetchFile(c.Path)
+			if err != nil && statusCode != 304 {
 				log.Error(err)
 				continue
 			}
 
-			c.lastUpdate = time.Now().UTC()
-			tCurrentUpdate := time.Now().UTC().Unix()
-
-			tmpVrpMap := make(map[string]*VRPJsonSimple)
-			for _, vrp := range decoded.Data {
-				asn, err := vrp.GetASN2()
+			var updatedVrpMap VRPMap
+			var inGracePeriod int
+			tCurrentUpdate := time.Now().UTC()
+			if statusCode == 304 {
+				updatedVrpMap, inGracePeriod = UpdateCurrentVrpMap(log.WithField("client", c.id), c.vrps, tCurrentUpdate)
+			} else {
+				log.Debug(data)
+				decoded, err := decodeJSON(data)
 				if err != nil {
-					log.Errorf("%d: exploration error for %v asn: %v", id, vrp, err)
-					continue
-				}
-				prefix, err := vrp.GetPrefix2()
-				if err != nil {
-					log.Errorf("%d: exploration error for %v prefix: %v", id, vrp, err)
+					log.Error(err)
 					continue
 				}
 
-				maxlen := vrp.GetMaxLen()
-				key := fmt.Sprintf("%s-%d-%d", prefix.String(), maxlen, asn)
-
-				firstSeen := tCurrentUpdate
-				currentEntry, ok := c.vrps[key]
-				if ok {
-					firstSeen = currentEntry.FirstSeen
-				}
-
-				vrpSimple := VRPJsonSimple{
-					Prefix:    prefix.String(),
-					ASN:       asn,
-					Length:    uint8(maxlen),
-					FirstSeen: firstSeen,
-				}
-				tmpVrpMap[key] = &vrpSimple
+				updatedVrpMap, inGracePeriod = BuildNewVrpMap(log.WithField("client", c.id), c.vrps, decoded.Data, tCurrentUpdate)
 			}
+
+			VRPInGracePeriod.With(prometheus.Labels{"url": c.Path}).Set(float64(inGracePeriod))
+
 			c.compLock.Lock()
-			c.vrps = tmpVrpMap
-			c.lastUpdate = time.Now().UTC()
+			c.vrps = updatedVrpMap
+			c.lastUpdate = tCurrentUpdate
 			c.compLock.Unlock()
 			if ch != nil {
 				ch <- id
@@ -371,6 +363,86 @@ func (c *Client) Start(id int, ch chan int) {
 
 }
 
+// Build the new vrpMap
+// The result:
+//   * contains all the VRPs in newVRPs
+//   * keeps the firstSeen value for VRPs already in the old map
+//   * keeps elements around for GracePeriod after they are not in the input.
+func BuildNewVrpMap(log *log.Entry, currentVrps VRPMap, newVrps []prefixfile.VRPJson, now time.Time) (VRPMap, int) {
+	tCurrentUpdate := now.Unix()
+	res := make(VRPMap)
+
+	for _, vrp := range newVrps {
+		asn, err := vrp.GetASN2()
+		if err != nil {
+			log.Errorf("exploration error for %v asn: %v", vrp, err)
+			continue
+		}
+		prefix, err := vrp.GetPrefix2()
+		if err != nil {
+			log.Errorf("exploration error for %v prefix: %v", vrp, err)
+			continue
+		}
+
+		maxlen := vrp.GetMaxLen()
+		key := fmt.Sprintf("%s-%d-%d", prefix.String(), maxlen, asn)
+
+		firstSeen := tCurrentUpdate
+		currentEntry, ok := currentVrps[key]
+		if ok && currentEntry.Visible {
+			// VRP is still visible, so keep the existing `FirstSeen`.
+			firstSeen = currentEntry.FirstSeen
+		}
+
+		res[key] = &VRPJsonSimple{
+			Prefix:    prefix.String(),
+			ASN:       asn,
+			Length:    uint8(maxlen),
+			FirstSeen: firstSeen,
+			LastSeen:  tCurrentUpdate,
+			Visible:   true,
+		}
+	}
+
+	// Copy objects that are within the grace period to the new map
+	gracePeriodEnds := tCurrentUpdate - int64(GracePeriod.Seconds())
+	inGracePeriod := 0
+	for k, entry := range currentVrps {
+		if _, ok := res[k]; !ok { // no longer present
+			if entry.LastSeen >= gracePeriodEnds {
+				entry.Visible = false
+				res[k] = entry
+				inGracePeriod++
+			}
+		}
+	}
+
+	return res, inGracePeriod
+}
+
+func UpdateCurrentVrpMap(log *log.Entry, currentVrps VRPMap, now time.Time) (VRPMap, int) {
+	tCurrentUpdate := now.Unix()
+	gracePeriodEnds := tCurrentUpdate - int64(GracePeriod.Seconds())
+	inGracePeriod := 0
+	res := make(VRPMap)
+
+	for key, vrp := range currentVrps {
+		if !vrp.Visible && vrp.LastSeen < gracePeriodEnds {
+			continue
+		}
+
+		copy := *vrp
+		if copy.Visible {
+			copy.LastSeen = tCurrentUpdate
+		} else {
+			inGracePeriod++
+		}
+		res[key] = &copy
+	}
+
+	return res, inGracePeriod
+}
+
 func (c *Client) HandlePDU(cs *rtr.ClientSession, pdu rtr.PDU) {
 	switch pdu := pdu.(type) {
 	case *rtr.PDUIPv4Prefix:
@@ -379,6 +451,7 @@ func (c *Client) HandlePDU(cs *rtr.ClientSession, pdu rtr.PDU) {
 			ASN:       pdu.ASN,
 			Length:    pdu.MaxLen,
 			FirstSeen: time.Now().Unix(),
+			Visible:   true,
 		}
 
 		key := fmt.Sprintf("%s-%d-%d", pdu.Prefix.String(), pdu.MaxLen, pdu.ASN)
@@ -397,6 +470,7 @@ func (c *Client) HandlePDU(cs *rtr.ClientSession, pdu rtr.PDU) {
 			ASN:       pdu.ASN,
 			Length:    pdu.MaxLen,
 			FirstSeen: time.Now().Unix(),
+			Visible:   true,
 		}
 
 		key := fmt.Sprintf("%s-%d-%d", pdu.Prefix.String(), pdu.MaxLen, pdu.ASN)
@@ -414,7 +488,7 @@ func (c *Client) HandlePDU(cs *rtr.ClientSession, pdu rtr.PDU) {
 
 		c.compRtrLock.Lock()
 		c.serial = pdu.SerialNumber
-		tmpVrpMap := make(map[string]*VRPJsonSimple, len(c.vrpsRtr))
+		tmpVrpMap := make(VRPMap, len(c.vrpsRtr))
 		for key, vrp := range c.vrpsRtr {
 			tmpVrpMap[key] = vrp
 		}
@@ -495,8 +569,9 @@ func (c *Client) continuousRTR(cs *rtr.ClientSession) {
 	}
 }
 
-func (c *Client) GetData() (map[string]*VRPJsonSimple, *diffMetadata) {
+func (c *Client) GetData() (VRPMap, *diffMetadata) {
 	c.compLock.RLock()
+	defer c.compLock.RUnlock()
 	vrps := c.vrps
 
 	md := &diffMetadata{
@@ -512,8 +587,6 @@ func (c *Client) GetData() (map[string]*VRPJsonSimple, *diffMetadata) {
 		LastFetch: c.lastUpdate.UnixNano() / 1e9,
 	}
 
-	c.compLock.RUnlock()
-
 	return vrps, md
 }
 
@@ -526,7 +599,7 @@ type Comparator struct {
 	OneOff bool
 
 	diffLock         *sync.RWMutex
-	onlyIn1, onlyIn2 []*VRPJsonSimple
+	onlyIn1, onlyIn2 VRPMap
 	md1              *diffMetadata
 	md2              *diffMetadata
 }
@@ -543,28 +616,55 @@ func NewComparator(c1, c2 *Client) *Comparator {
 	}
 }
 
-func countFirstSeenOnOrBefore(vrps []*VRPJsonSimple, thresholdTimestamp int64) float64 {
+func isDefinitelyVisible(vrp *VRPJsonSimple, thresholdTimestamp int64) bool {
+	return vrp != nil && vrp.Visible && vrp.FirstSeen <= thresholdTimestamp
+}
+
+func isDefinitelyNotVisisble(vrp *VRPJsonSimple, thresholdTimestamp int64) bool {
+	return vrp == nil || (!vrp.Visible && vrp.LastSeen <= thresholdTimestamp)
+}
+
+func countUnmatched(vrps VRPMap, others VRPMap, thresholdTimestamp int64) float64 {
 	count := 0
 
-	for _, vrp := range vrps {
-		if vrp.FirstSeen <= thresholdTimestamp {
-			count++
+	for key, vrp := range vrps {
+		other := others[key]
+		if isDefinitelyVisible(vrp, thresholdTimestamp) {
+			// VRP is definitely visible, count if other side is definitely not visible
+			if isDefinitelyNotVisisble(other, thresholdTimestamp) {
+				count++
+			}
+		} else if isDefinitelyNotVisisble(vrp, thresholdTimestamp) {
+			// VRP is definitely not visible, count if other side is definitely visible
+			if isDefinitelyVisible(other, thresholdTimestamp) {
+				count++
+			}
 		}
 	}
 
 	return float64(count)
 }
 
-func Diff(a, b map[string]*VRPJsonSimple) []*VRPJsonSimple {
-	onlyInA := make([]*VRPJsonSimple, 0)
-	for key, vrp := range a {
-		if _, ok := b[key]; !ok {
-			onlyInA = append(onlyInA, vrp)
+func Diff(a, b VRPMap) VRPMap {
+	onlyInA := make(VRPMap)
+	for key, vrpA := range a {
+		if vrpA.Visible {
+			if vrpB, ok := b[key]; !ok || !vrpB.Visible {
+				onlyInA[key] = vrpA
+			}
 		}
 	}
 	return onlyInA
 }
 
+func VRPArray(a VRPMap) []*VRPJsonSimple {
+	result := make([]*VRPJsonSimple, 0)
+	for _, vrp := range a {
+		result = append(result, vrp)
+	}
+	return result
+}
+
 type diffMetadata struct {
 	LastFetch int64  `json:"last-fetch"`
 	URL       string `json:"url"`
@@ -582,7 +682,10 @@ type VRPJsonSimple struct {
 	Length    uint8  `json:"max-length"`
 	Prefix    string `json:"prefix"`
 	FirstSeen int64  `json:"first-seen"`
+	LastSeen  int64  `json:"last-seen"`
+	Visible   bool   `json:"visible"`
 }
+type VRPMap map[string]*VRPJsonSimple
 
 type diffExport struct {
 	MetadataPrimary   *diffMetadata    `json:"metadata-primary"`
@@ -604,8 +707,8 @@ func (c *Comparator) ServeDiff(wr http.ResponseWriter, req *http.Request) {
 	export := diffExport{
 		MetadataPrimary:   md1,
 		MetadataSecondary: md2,
-		OnlyInPrimary:     d1,
-		OnlyInSecondary:   d2,
+		OnlyInPrimary:     VRPArray(d1),
+		OnlyInSecondary:   VRPArray(d2),
 	}
 
 	wr.Header().Add("content-type", "application/json")
@@ -676,14 +779,14 @@ func (c *Comparator) Compare() {
 							"lhs_url":            md1.URL,
 							"rhs_url":            md2.URL,
 							"visibility_seconds": strconv.FormatInt(visibleFor, 10),
-						}).Set(countFirstSeenOnOrBefore(onlyIn1, thresholdTimestamp))
+						}).Set(countUnmatched(onlyIn1, vrps2, thresholdTimestamp))
 
 					VRPDifferenceForDuration.With(
 						prometheus.Labels{
 							"lhs_url":            md2.URL,
 							"rhs_url":            md1.URL,
 							"visibility_seconds": strconv.FormatInt(visibleFor, 10),
-						}).Set(countFirstSeenOnOrBefore(onlyIn2, thresholdTimestamp))
+						}).Set(countUnmatched(onlyIn2, vrps1, thresholdTimestamp))
 				}
 			}
 
@@ -765,8 +868,6 @@ func (c *Comparator) Start() error {
 }
 
 func main() {
-	runtime.GOMAXPROCS(runtime.NumCPU())
-
 	flag.Parse()
 	if flag.NArg() > 0 {
 		fmt.Printf("%s: illegal positional argument(s) provided (\"%s\") - did you mean to provide a flag?\n", os.Args[0], strings.Join(flag.Args(), " "))
@@ -780,6 +881,12 @@ func main() {
 	lvl, _ := log.ParseLevel(*LogLevel)
 	log.SetLevel(lvl)
 
+	highestVisibilityThreshold := time.Second * time.Duration(visibilityThresholds[len(visibilityThresholds)-1])
+	if highestVisibilityThreshold > *GracePeriod {
+		log.Warnf("Highest visibility threshold %v greater than grace period %v, adjusting grace period", highestVisibilityThreshold, GracePeriod)
+		*GracePeriod = highestVisibilityThreshold
+	}
+
 	fc := utils.NewFetchConfig()
 	fc.EnableEtags = !*DisableConditionalRequests
 	fc.EnableLastModified = !*DisableConditionalRequests
@@ -810,7 +917,7 @@ func main() {
 			keyBytesStr := os.Getenv(fmt.Sprintf("%s_1", ENV_SSH_KEY))
 			keyBytes = []byte(keyBytesStr)
 		} else {
-			keyBytes, err = ioutil.ReadFile(*PrimarySSHAuthKey)
+			keyBytes, err = os.ReadFile(*PrimarySSHAuthKey)
 			if err != nil {
 				log.Fatal(err)
 			}
@@ -842,7 +949,7 @@ func main() {
 			keyBytesStr := os.Getenv(fmt.Sprintf("%s_2", ENV_SSH_KEY))
 			keyBytes = []byte(keyBytesStr)
 		} else {
-			keyBytes, err = ioutil.ReadFile(*SecondarySSHAuthKey)
+			keyBytes, err = os.ReadFile(*SecondarySSHAuthKey)
 			if err != nil {
 				log.Fatal(err)
 			}
@@ -855,6 +962,7 @@ func main() {
 	go func() {
 		http.HandleFunc(fmt.Sprintf("/%s", *OutFile), cmp.ServeDiff)
 		http.Handle(*MetricsPath, promhttp.Handler())
+		http.HandleFunc("/", ServeIndex)
 
 		log.Fatal(http.ListenAndServe(*Addr, nil))
 	}()
@@ -862,3 +970,20 @@ func main() {
 	log.Fatal(cmp.Start())
 
 }
+
+type IndexTemplateVars struct {
+	MetricsPath string
+	OutFile     string
+	Addr        string
+}
+
+func ServeIndex(wr http.ResponseWriter, req *http.Request) {
+	tmpl, err := template.New("index").Parse(IndexTemplate)
+	if err == nil {
+		err = tmpl.Execute(wr, IndexTemplateVars{*MetricsPath, *OutFile, *Addr})
+	}
+
+	if err != nil {
+		io.WriteString(wr, IndexTemplate)
+	}
+}
