@@ -9,12 +9,10 @@ import (
 	"errors"
 	"flag"
 	"fmt"
-	"io/ioutil"
 	"net"
 	"net/http"
 	"os"
 	"os/signal"
-	"runtime"
 	"strings"
 	"sync"
 	"syscall"
@@ -47,7 +45,7 @@ var (
 	buildinfos = ""
 	AppVersion = "StayRTR " + version + " " + buildinfos
 
-	MetricsAddr = flag.String("metrics.addr", ":8080", "Metrics address")
+	MetricsAddr = flag.String("metrics.addr", ":9847", "Metrics address")
 	MetricsPath = flag.String("metrics.path", "/metrics", "Metrics path")
 
 	ExportPath = flag.String("export.path", "/rpki.json", "Export path")
@@ -167,9 +165,11 @@ func metricHTTP() {
 	log.Fatal(http.ListenAndServe(*MetricsAddr, nil))
 }
 
-func checkFile(data []byte) ([]byte, error) {
-	hsum := sha256.Sum256(data)
-	return hsum[:], nil
+// newSHA256 will return the sha256 sum of the byte slice
+// The return will be converted form a [32]byte to []byte
+func newSHA256(data []byte) []byte {
+	hash := sha256.Sum256(data)
+	return hash[:]
 }
 
 func decodeJSON(data []byte) (*prefixfile.VRPList, error) {
@@ -181,24 +181,35 @@ func decodeJSON(data []byte) (*prefixfile.VRPList, error) {
 	return &vrplistjson, err
 }
 
-func checkPrefixLengths(prefix *net.IPNet, maxLength uint8) bool {
+func isValidPrefixLength(prefix *net.IPNet, maxLength uint8) bool {
 	plen, max := net.IPMask.Size(prefix.Mask)
 
-	if uint8(plen) > maxLength || maxLength > uint8(max) {
+	if plen == 0 || uint8(plen) > maxLength || maxLength > uint8(max) {
 		log.Errorf("%s Maxlength wrong: %d - %d", prefix, plen, maxLength)
 		return false
 	}
 	return true
 }
 
+// processData will take a slice of prefix.VRPJson and attempt to convert them to a slice of rtr.VRP.
+// Will check the following:
+// 1 - The prefix is a valid prefix
+// 2 - The ASN is a valid ASN
+// 3 - The MaxLength is valid
+// Will return a deduped slice, as well as total VRPs, IPv4 VRPs, and IPv6 VRPs
 func processData(vrplistjson []prefixfile.VRPJson) ([]rtr.VRP, int, int, int) {
 	filterDuplicates := make(map[string]bool)
 
-	vrplist := make([]rtr.VRP, 0)
+	// It may be tempting to change this to a simple time.Since() but that will
+	// grab the current time every time it's invoked, time calls can be slow on
+	// some platforms, so lets just get the unix time when we start and use that
+	// to compare it all
+	NowUnix := time.Now().UTC().Unix()
 
-	var count int
+	var vrplist []rtr.VRP
 	var countv4 int
 	var countv6 int
+
 	for _, v := range vrplistjson {
 		prefix, err := v.GetPrefix2()
 		if err != nil {
@@ -211,24 +222,30 @@ func processData(vrplistjson []prefixfile.VRPJson) ([]rtr.VRP, int, int, int) {
 			continue
 		}
 
-		if !checkPrefixLengths(prefix, v.Length) {
+		if !isValidPrefixLength(prefix, v.Length) {
 			continue
 		}
 
+		if v.Expires != nil {
+			// Prevent stale VRPs from being considered
+			// https://github.com/bgp/stayrtr/issues/15
+			if int(NowUnix) > int(*v.Expires) {
+				continue
+			}
+		}
+
 		if prefix.IP.To4() != nil {
 			countv4++
-		} else if prefix.IP.To16() != nil {
+		} else {
 			countv6++
 		}
-		count++
 
 		key := fmt.Sprintf("%s,%d,%d", prefix, asn, v.Length)
 		_, exists := filterDuplicates[key]
-		if !exists {
-			filterDuplicates[key] = true
-		} else {
+		if exists {
 			continue
 		}
+		filterDuplicates[key] = true
 
 		vrp := rtr.VRP{
 			Prefix: *prefix,
@@ -237,7 +254,7 @@ func processData(vrplistjson []prefixfile.VRPJson) ([]rtr.VRP, int, int, int) {
 		}
 		vrplist = append(vrplist, vrp)
 	}
-	return vrplist, count, countv4, countv6
+	return vrplist, countv4 + countv6, countv4, countv6
 }
 
 type IdenticalFile struct {
@@ -248,22 +265,29 @@ func (e IdenticalFile) Error() string {
 	return fmt.Sprintf("File %s is identical to the previous version", e.File)
 }
 
+var errVRPJsonFileTooOld = errors.New("VRP JSON file is older than 24 hours")
+
 // Update the state based on the current slurm file and data.
 func (s *state) updateFromNewState() error {
-	sessid, _ := s.server.GetSessionId(nil)
+	sessid := s.server.GetSessionId()
 
+	vrpsjson := s.lastdata.Data
+	if vrpsjson == nil {
+		return nil
+	}
+
+	buildtime, err := time.Parse(time.RFC3339, s.lastdata.Metadata.Buildtime)
 	if s.checktime {
-		buildtime, err := time.Parse(time.RFC3339, s.lastdata.Metadata.Buildtime)
 		if err != nil {
 			return err
 		}
 		notafter := buildtime.Add(time.Hour * 24)
 		if time.Now().UTC().After(notafter) {
-			return errors.New(fmt.Sprintf("VRP JSON file is older than 24 hours: %v", buildtime))
+			log.Warnf("VRP JSON file is older than 24 hours: %v", buildtime)
+			return errVRPJsonFileTooOld
 		}
 	}
 
-	vrpsjson := s.lastdata.Data
 	if s.slurm != nil {
 		kept, removed := s.slurm.FilterOnVRPs(vrpsjson)
 		asserted := s.slurm.AssertVRPs()
@@ -273,9 +297,47 @@ func (s *state) updateFromNewState() error {
 
 	vrps, count, countv4, countv6 := processData(vrpsjson)
 
-	log.Infof("New update (%v uniques, %v total prefixes).",
-		len(vrps), count)
+	log.Infof("New update (%v uniques, %v total prefixes).", len(vrps), count)
+	return s.applyUpdateFromNewState(vrps, sessid, vrpsjson, countv4, countv6)
+}
 
+// Update the state based on the currently loaded files
+func (s *state) reloadFromCurrentState() error {
+	sessid := s.server.GetSessionId()
+
+	vrpsjson := s.lastdata.Data
+	if vrpsjson == nil {
+		return nil
+	}
+
+	buildtime, err := time.Parse(time.RFC3339, s.lastdata.Metadata.Buildtime)
+	if s.checktime {
+		if err != nil {
+			return err
+		}
+		notafter := buildtime.Add(time.Hour * 24)
+		if time.Now().UTC().After(notafter) {
+			log.Warnf("VRP JSON file is older than 24 hours: %v", buildtime)
+			return errVRPJsonFileTooOld
+		}
+	}
+
+	if s.slurm != nil {
+		kept, removed := s.slurm.FilterOnVRPs(vrpsjson)
+		asserted := s.slurm.AssertVRPs()
+		log.Infof("Slurm filtering: %v kept, %v removed, %v asserted", len(kept), len(removed), len(asserted))
+		vrpsjson = append(kept, asserted...)
+	}
+
+	vrps, count, countv4, countv6 := processData(vrpsjson)
+	if s.server.CountVRPs() != count {
+		log.Infof("New update to old state (%v uniques, %v total prefixes). (old %v - new %v)", len(vrps), count, s.server.CountVRPs(), count)
+		return s.applyUpdateFromNewState(vrps, sessid, vrpsjson, countv4, countv6)
+	}
+	return nil
+}
+
+func (s *state) applyUpdateFromNewState(vrps []rtr.VRP, sessid uint16, vrpsjson []prefixfile.VRPJson, countv4 int, countv6 int) error {
 	s.server.AddVRPs(vrps)
 
 	serial, _ := s.server.GetCurrentSerial(sessid)
@@ -327,7 +389,7 @@ func (s *state) updateFile(file string) (bool, error) {
 		RefreshStatusCode.WithLabelValues(file, fmt.Sprintf("%d", code)).Inc()
 	}
 
-	hsum, _ := checkFile(data)
+	hsum := newSHA256(data)
 	if s.lasthash != nil {
 		cres := bytes.Compare(s.lasthash, hsum)
 		if cres == 0 {
@@ -376,50 +438,91 @@ func (s *state) routineUpdate(file string, interval int, slurmFile string) {
 	log.Debugf("Starting refresh routine (file: %v, interval: %vs, slurm: %v)", file, interval, slurmFile)
 	signals := make(chan os.Signal, 1)
 	signal.Notify(signals, syscall.SIGHUP)
+	delay := time.NewTicker(time.Duration(interval) * time.Second)
+	initialSyncNotComplete := false
 	for {
-		delay := time.NewTimer(time.Duration(interval) * time.Second)
+		if s.lastchange.IsZero() {
+			log.Warn("Initial sync not complete. Refreshing every 30 seconds")
+			delay.Reset(30 * time.Second)
+			initialSyncNotComplete = true
+		} else {
+			if initialSyncNotComplete {
+				delay.Reset(time.Duration(interval) * time.Second)
+			}
+		}
 		select {
 		case <-delay.C:
 		case <-signals:
 			log.Debug("Received HUP signal")
+			if s.lastchange.IsZero() {
+				delay.Reset(30 * time.Second)
+			} else {
+				delay.Reset(time.Duration(interval) * time.Second)
+			}
 		}
-		delay.Stop()
 		slurmNotPresentOrUpdated := false
-		if slurmFile != "" {
+
+		updateFileWG := sync.WaitGroup{}
+		updateFileWG.Add(2)
+		go func() {
+			defer updateFileWG.Done()
+			if slurmFile != "" {
+				var err error
+				slurmNotPresentOrUpdated, err = s.updateSlurm(slurmFile)
+				if err != nil {
+					switch err.(type) {
+					case utils.HttpNotModified:
+						log.Info(err)
+					case utils.IdenticalEtag:
+						log.Info(err)
+					default:
+						log.Errorf("Slurm: %v", err)
+					}
+				}
+			}
+		}()
+		var cacheUpdated bool
+
+		go func() {
+			defer updateFileWG.Done()
 			var err error
-			slurmNotPresentOrUpdated, err = s.updateSlurm(slurmFile)
+			cacheUpdated, err = s.updateFile(file)
 			if err != nil {
 				switch err.(type) {
 				case utils.HttpNotModified:
 					log.Info(err)
 				case utils.IdenticalEtag:
 					log.Info(err)
+				case IdenticalFile:
+					log.Info(err)
 				default:
-					log.Errorf("Slurm: %v", err)
+					log.Errorf("Error updating: %v", err)
 				}
 			}
-		}
-		cacheUpdated, err := s.updateFile(file)
-		if err != nil {
-			switch err.(type) {
-			case utils.HttpNotModified:
-				log.Info(err)
-			case utils.IdenticalEtag:
-				log.Info(err)
-			case IdenticalFile:
-				log.Info(err)
-			default:
-				log.Errorf("Error updating: %v", err)
-			}
-		}
+		}()
+
+		updateFileWG.Wait()
 
 		// Only process the first time after there is either a cache or SLURM
 		// update.
 		if cacheUpdated || slurmNotPresentOrUpdated {
 			err := s.updateFromNewState()
 			if err != nil {
+				if err == errVRPJsonFileTooOld {
+					// If the exiting build time is over 24 hours, It's time to drop everything out.
+					// to avoid routing on stale data
+					buildTime := s.exported.Metadata.GetBuildTime()
+					if !buildTime.IsZero() && time.Since(buildTime) > time.Hour*24 {
+						s.server.AddVRPs([]rtr.VRP{}) // empty the store of VRP by giving it a empty VRP array, triggering a emptying
+					}
+				}
 				log.Errorf("Error updating from new state: %v", err)
 			}
+		} else {
+			err := s.reloadFromCurrentState()
+			if err != nil {
+				log.Errorf("Error updating state: %v", err)
+			}
 		}
 	}
 }
@@ -484,8 +587,13 @@ func (m *metricsEvent) UpdateMetrics(numIPv4 int, numIPv6 int, numIPv4filtered i
 }
 
 func main() {
-	runtime.GOMAXPROCS(runtime.NumCPU())
+	if err := run(); err != nil {
+		fmt.Fprintf(os.Stderr, "%s\n", err)
+		os.Exit(1)
+	}
+}
 
+func run() error {
 	flag.Parse()
 	if flag.NArg() > 0 {
 		fmt.Printf("%s: illegal positional argument(s) provided (\"%s\") - did you mean to provide a flag?\n", os.Args[0], strings.Join(flag.Args(), " "))
@@ -528,6 +636,7 @@ func main() {
 
 	s := state{
 		server:       server,
+		lastdata:     &prefixfile.VRPList{},
 		metricsEvent: me,
 		sendNotifs:   *SendNotifs,
 		checktime:    *TimeCheck,
@@ -548,47 +657,61 @@ func main() {
 	}
 
 	if *Bind == "" && *BindTLS == "" && *BindSSH == "" {
-		log.Fatalf("Specify at least a bind address")
+		log.Fatalf("Specify at least a bind address using -bind , -tls.bind , or -ssh.bind")
 	}
 
-	_, err := s.updateFile(*CacheBin)
-	if err != nil {
-		switch err.(type) {
-		case utils.HttpNotModified:
-			log.Info(err)
-		case IdenticalFile:
-			log.Info(err)
-		case utils.IdenticalEtag:
-			log.Info(err)
-		default:
-			log.Errorf("Error updating: %v", err)
-		}
-	}
+	fileFetchWG := sync.WaitGroup{}
+	fileFetchWG.Add(2)
 
-	slurmFile := *Slurm
-	if slurmFile != "" {
-		_, err := s.updateSlurm(slurmFile)
+	go func() {
+		defer fileFetchWG.Done()
+		_, err := s.updateFile(*CacheBin)
 		if err != nil {
 			switch err.(type) {
 			case utils.HttpNotModified:
 				log.Info(err)
+			case IdenticalFile:
+				log.Info(err)
 			case utils.IdenticalEtag:
 				log.Info(err)
 			default:
-				log.Errorf("Slurm: %v", err)
+				log.Errorf("Error updating: %v", err)
 			}
 		}
-		if !*SlurmRefresh {
-			slurmFile = ""
+	}()
+
+	slurmFile := *Slurm
+	go func() {
+		defer fileFetchWG.Done()
+		if slurmFile != "" {
+			_, err := s.updateSlurm(slurmFile)
+			if err != nil {
+				switch err.(type) {
+				case utils.HttpNotModified:
+					log.Info(err)
+				case utils.IdenticalEtag:
+					log.Info(err)
+				default:
+					log.Errorf("Slurm: %v", err)
+				}
+			}
+			if !*SlurmRefresh {
+				slurmFile = ""
+			}
 		}
-	}
+	}()
+
+	fileFetchWG.Wait()
 
 	// Initial calculation of state (after fetching cache + slurm)
-	s.updateFromNewState()
+	err := s.updateFromNewState()
+	if err != nil {
+		log.Warnf("Error setting up initial state: %s", err)
+	}
 
 	if *Bind != "" {
 		go func() {
-			sessid, _ := server.GetSessionId(nil)
+			sessid := server.GetSessionId()
 			log.Infof("StayRTR Server started (sessionID:%d, refresh:%d, retry:%d, expire:%d)", sessid, sc.RefreshInterval, sc.RetryInterval, sc.ExpireInterval)
 			err := server.Start(*Bind)
 			if err != nil {
@@ -612,13 +735,13 @@ func main() {
 		}()
 	}
 	if *BindSSH != "" {
-		sshkey, err := ioutil.ReadFile(*SSHKey)
+		sshkey, err := os.ReadFile(*SSHKey)
 		if err != nil {
 			log.Fatal(err)
 		}
 		private, err := ssh.ParsePrivateKey(sshkey)
 		if err != nil {
-			log.Fatal("Failed to parse private key: ", err)
+			log.Fatal("Failed to parse SSH private key: ", err)
 		}
 
 		sshConfig := ssh.ServerConfig{}
@@ -629,11 +752,11 @@ func main() {
 			if password == "" {
 				password = os.Getenv(ENV_SSH_PASSWORD)
 			}
-			sshConfig.PasswordCallback = func(conn ssh.ConnMetadata, password []byte) (*ssh.Permissions, error) {
+			sshConfig.PasswordCallback = func(conn ssh.ConnMetadata, suppliedPassword []byte) (*ssh.Permissions, error) {
 				log.Infof("Connected (ssh-password): %v/%v", conn.User(), conn.RemoteAddr())
-				if conn.User() != *SSHAuthUser || !bytes.Equal(password, []byte(*SSHAuthPassword)) {
+				if conn.User() != *SSHAuthUser || !bytes.Equal(suppliedPassword, []byte(password)) {
 					log.Warnf("Wrong user or password for %v/%v. Disconnecting.", conn.User(), conn.RemoteAddr())
-					return nil, errors.New("Wrong user or password")
+					return nil, errors.New("wrong user or password")
 				}
 
 				return &ssh.Permissions{
@@ -647,7 +770,7 @@ func main() {
 			if *SSHAuthKeysList == "" {
 				sshClientKeysToDecode = os.Getenv(ENV_SSH_KEY)
 			} else {
-				sshClientKeysToDecodeBytes, err := ioutil.ReadFile(*SSHAuthKeysList)
+				sshClientKeysToDecodeBytes, err := os.ReadFile(*SSHAuthKeysList)
 				if err != nil {
 					log.Fatal(err)
 				}
@@ -663,7 +786,7 @@ func main() {
 						if k == "" {
 							continue
 						}
-						if strings.HasPrefix(fmt.Sprintf("%v %v", key.Type(), keyBase64), k) {
+						if strings.HasPrefix(k, fmt.Sprintf("%v %v", key.Type(), keyBase64)) {
 							log.Infof("Connected (ssh-key): %v/%v with key %v %v (matched with line %v)",
 								conn.User(), conn.RemoteAddr(), key.Type(), keyBase64, i+1)
 							noKeys = true
@@ -672,7 +795,7 @@ func main() {
 					}
 					if !noKeys {
 						log.Warnf("No key for %v/%v %v %v. Disconnecting.", conn.User(), conn.RemoteAddr(), key.Type(), keyBase64)
-						return nil, errors.New("Key not found")
+						return nil, errors.New("provided ssh key not found")
 					}
 				} else {
 					log.Infof("Connected (ssh-key): %v/%v with key %v %v", conn.User(), conn.RemoteAddr(), key.Type(), keyBase64)
@@ -700,4 +823,5 @@ func main() {
 
 	s.routineUpdate(*CacheBin, *RefreshInterval, slurmFile)
 
+	return nil
 }
