@@ -4,12 +4,13 @@ import (
 	"bytes"
 	"crypto/tls"
 	"fmt"
-	"golang.org/x/crypto/ssh"
 	"io"
 	"math/rand"
 	"net"
 	"sync"
 	"time"
+
+	"golang.org/x/crypto/ssh"
 )
 
 func GenerateSessionId() uint16 {
@@ -32,7 +33,7 @@ type RTREventHandler interface {
 
 type VRPManager interface {
 	GetCurrentSerial(uint16) (uint32, bool)
-	GetSessionId(*Client) (uint16, error)
+	GetSessionId() uint16
 	GetCurrentVRPs() ([]VRP, bool)
 	GetVRPsSerialDiff(uint32) ([]VRP, bool)
 }
@@ -50,7 +51,7 @@ func (e *DefaultRTREventHandler) RequestCache(c *Client) {
 	if e.Log != nil {
 		e.Log.Debugf("%v > Request Cache", c)
 	}
-	sessionId, _ := e.vrpManager.GetSessionId(c)
+	sessionId := e.vrpManager.GetSessionId()
 	serial, valid := e.vrpManager.GetCurrentSerial(sessionId)
 	if !valid {
 		c.SendNoDataError()
@@ -273,8 +274,8 @@ func (s *Server) SetManualSerial(v bool) {
 	s.manualserial = v
 }
 
-func (s *Server) GetSessionId(c *Client) (uint16, error) {
-	return s.sessId, nil
+func (s *Server) GetSessionId() uint16 {
+	return s.sessId
 }
 
 func (s *Server) GetCurrentVRPs() ([]VRP, bool) {
@@ -338,14 +339,20 @@ func (s *Server) setSerial(serial uint32) {
 // be called before the VRPs data is added.
 func (s *Server) SetSerial(serial uint32) {
 	s.vrplock.RLock()
+	defer s.vrplock.RUnlock()
 	//s.vrpListSerial = make([]uint32, 0)
 	s.setSerial(serial)
-	s.vrplock.RUnlock()
+}
+
+func (s *Server) CountVRPs() int {
+	s.vrplock.RLock()
+	defer s.vrplock.RUnlock()
+
+	return len(s.vrpCurrent)
 }
 
 func (s *Server) AddVRPs(vrps []VRP) {
 	s.vrplock.RLock()
-	curDiff := make([]VRP, 0)
 
 	vrpCurrent := s.vrpCurrent
 
@@ -355,7 +362,7 @@ func (s *Server) AddVRPs(vrps []VRP) {
 	} else if s.log != nil {
 		s.log.Debugf("Computed diff: added (%d), removed (%d), unchanged (%d)", len(added), len(removed), len(unchanged))
 	}
-	curDiff = append(added, removed...)
+	curDiff := append(added, removed...)
 	s.vrplock.RUnlock()
 
 	s.AddVRPsDiff(curDiff)
@@ -383,6 +390,7 @@ func (s *Server) AddVRPsDiff(diff []VRP) {
 	s.vrplock.RUnlock()
 
 	s.vrplock.Lock()
+	defer s.vrplock.Unlock()
 	newserial := s.generateSerial()
 	removed := s.addSerial(newserial)
 
@@ -407,7 +415,6 @@ func (s *Server) AddVRPsDiff(diff []VRP) {
 	s.vrpListDiff = nextDiff
 	s.vrpCurrent = newVrpCurrent
 	s.setSerial(newserial)
-	s.vrplock.Unlock()
 }
 
 func (s *Server) SetBaseVersion(version uint8) {
@@ -614,7 +621,6 @@ func (s *Server) loopTCP(tcplist net.Listener, logEnv string, clientCallback Cli
 			}
 		}
 	}
-	return nil
 }
 
 func (s *Server) StartSSH(bind string, config *ssh.ServerConfig) error {
@@ -851,7 +857,7 @@ func (r VRP) String() string {
 }
 
 func (r1 VRP) Equals(r2 VRP) bool {
-	return r1.MaxLen == r2.MaxLen && r1.ASN == r2.ASN && bytes.Equal(r1.Prefix.IP, r2.Prefix.IP) && bytes.Equal(r1.Prefix.Mask, r2.Prefix.Mask)
+	return r1.MaxLen == r2.MaxLen && r1.ASN == r2.ASN && r1.Prefix.IP.Equal(r2.Prefix.IP) && bytes.Equal(r1.Prefix.Mask, r2.Prefix.Mask)
 }
 
 func (r1 VRP) Copy() VRP {
@@ -957,7 +963,6 @@ func (c *Client) Disconnect() {
 	select {
 	case c.quit <- true:
 	default:
-
 	}
 
 	c.tcpconn.Close()
